@@ -43,6 +43,8 @@ typedef struct
     GObjectClass parent_class;
 
     void (*image_changed) (Page *page);
+    void (*size_changed) (Page *page);
+    void (*scan_line_changed) (Page *page);
     void (*orientation_changed) (Page *page);
     void (*crop_changed) (Page *page);
 } PageClass;
@@ -69,6 +71,8 @@ gint page_get_scan_height (Page *page);
 
 void page_start (Page *page);
 
+gboolean page_has_data (Page *page);
+
 gint page_get_scan_line (Page *page);
 
 void page_parse_scan_line (Page *page, ScanLine *line);
@@ -103,4 +107,8 @@ GdkPixbuf *page_get_image (Page *page);
 
 GdkPixbuf *page_get_cropped_image (Page *page);
 
+gboolean page_save_jpeg (Page *page, const gchar *uri, GError **error);
+
+gboolean page_save_png (Page *page, const gchar *uri, GError **error);
+
 #endif /* _PAGE_H_ */
