@@ -16,15 +16,18 @@
 
 #include "scanner.h"
 
+/* TODO: Could indicate the start of the next page immediately after the last page is received (i.e. before the sane_cancel()) */
 
 enum {
     UPDATE_DEVICES,
     AUTHORIZE,
+    EXPECT_PAGE,
     GOT_PAGE_INFO,
     GOT_LINE,
     SCAN_FAILED,
     PAGE_DONE,
     DOCUMENT_DONE,
+    SCANNING_CHANGED,
     LAST_SIGNAL
 };
 static guint signals[LAST_SIGNAL] = { 0, };
@@ -78,7 +81,7 @@ struct ScannerPrivate
     GAsyncQueue *scan_queue, *authorize_queue;
     GThread *thread;
 
-    ScanDevice *default_device;
+    gchar *default_device;
 
     ScanState state;
     gboolean redetect;
@@ -96,9 +99,10 @@ struct ScannerPrivate
 
     /* Buffer for received line */
     SANE_Byte *buffer;
+    SANE_Int n_used;
 
     SANE_Int bytes_remaining, line_count, pass_number, page_number, notified_page;
-  
+
     gboolean scanning;
 };
 
@@ -108,15 +112,14 @@ G_DEFINE_TYPE (Scanner, scanner, G_TYPE_OBJECT);
 /* Table of scanner objects for each thread (required for authorization callback) */
 static GHashTable *scanners;
 
+#define MAX_READ 10000
+
 
 static gboolean
 send_signal (SignalInfo *info)
 {
     g_signal_emit (info->instance, signals[info->sig], 0, info->data);
 
-    if (info->sig == DOCUMENT_DONE || info->sig == SCAN_FAILED)
-        info->instance->priv->scanning = FALSE;
-
     switch (info->sig) {
     case UPDATE_DEVICES:
         {
@@ -156,6 +159,7 @@ send_signal (SignalInfo *info)
         }
         break;
     default:
+    case EXPECT_PAGE:
     case PAGE_DONE:
     case DOCUMENT_DONE:
     case LAST_SIGNAL:
@@ -182,16 +186,28 @@ emit_signal (Scanner *scanner, guint sig, gpointer data)
 }
 
 
+static void
+set_scanning (Scanner *scanner, gboolean is_scanning)
+{
+    if ((scanner->priv->scanning && !is_scanning) || (!scanner->priv->scanning && is_scanning)) {
+        scanner->priv->scanning = is_scanning;
+        emit_signal (scanner, SCANNING_CHANGED, NULL);
+    }
+}
+
+
 static gint get_device_weight (const gchar *device)
 {
-    /* Use locally connected devices first */
-    if (g_str_has_prefix (device, "usb:"))
-       return 0;
-
+    /* NOTE: This is using trends in the naming of SANE devices, SANE should be able to provide this information better */
+  
     /* Use webcams as a last resort */
     if (g_str_has_prefix (device, "vfl:"))
        return 2;
 
+    /* Use locally connected devices first */
+    if (strstr (device, "usb"))
+       return 0;
+
     return 1;
 }
 
@@ -201,7 +217,7 @@ compare_devices (ScanDevice *device1, ScanDevice *device2)
 {
     gint weight1, weight2;
   
-    /* Should do some fuzzy matching on the last selected device and set that to the default */
+    /* TODO: Should do some fuzzy matching on the last selected device and set that to the default */
 
     weight1 = get_device_weight (device1->name);
     weight2 = get_device_weight (device2->name);
@@ -322,8 +338,7 @@ do_redetect (Scanner *scanner)
     for (device_iter = device_list; *device_iter; device_iter++) {
         const SANE_Device *device = *device_iter;
         ScanDevice *scan_device;
-        GString *label;
-        gchar *c;
+        gchar *c, *label;
 
         g_debug ("Device: name=\"%s\" vendor=\"%s\" model=\"%s\" type=\"%s\"",
                  device->name, device->vendor, device->model, device->type);
@@ -332,20 +347,18 @@ do_redetect (Scanner *scanner)
 
         scan_device->name = g_strdup (device->name);
 
-        label = g_string_new ("");
         /* Abbreviate HP as it is a long string and does not match what is on the physical scanner */
         if (strcmp (device->vendor, "Hewlett-Packard") == 0)
-            g_string_printf (label, "HP %s", device->model);
+            label = g_strdup_printf ("HP %s", device->model);
         else
-            g_string_printf (label, "%s %s", device->vendor, device->model);
+            label = g_strdup_printf ("%s %s", device->vendor, device->model);
         
         /* Replace underscored in name */
-        for (c = label->str; *c; c++)
+        for (c = label; *c; c++)
             if (*c == '_')
                 *c = ' ';
 
-        scan_device->label = label->str;
-        g_string_free (label, FALSE);
+        scan_device->label = label;
 
         devices = g_list_append (devices, scan_device);
     }
@@ -356,8 +369,11 @@ do_redetect (Scanner *scanner)
     scanner->priv->redetect = FALSE;
     scanner->priv->state = STATE_IDLE;
 
-    if (devices)
-        scanner->priv->default_device = g_list_nth_data (devices, 0);
+    g_free (scanner->priv->default_device);
+    if (devices) {
+        ScanDevice *device = g_list_nth_data (devices, 0);
+        scanner->priv->default_device = g_strdup (device->name);
+    }
     else
         scanner->priv->default_device = NULL;
 
@@ -366,12 +382,12 @@ do_redetect (Scanner *scanner)
 
 
 static gboolean
-control_option (SANE_Handle handle, SANE_Value_Type type, SANE_Int index, SANE_Action action, void *value)
+control_option (SANE_Handle handle, const SANE_Option_Descriptor *option, SANE_Int index, SANE_Action action, void *value)
 {
     SANE_Status status;
     
     status = sane_control_option (handle, index, action, value, NULL);
-    switch (type) {
+    switch (option->type) {
     case SANE_TYPE_BOOL:
         g_debug ("sane_control_option (%d, %s, %s) -> %s",
                  index, get_action_string (action),
@@ -399,16 +415,16 @@ control_option (SANE_Handle handle, SANE_Value_Type type, SANE_Int index, SANE_A
     default:
         break;
     }
-  
+
     if (status != SANE_STATUS_GOOD)
-        g_warning ("Error setting control option: %s", sane_strstatus(status));
+        g_warning ("Error setting option %s: %s", option->name, sane_strstatus(status));
 
     return status == SANE_STATUS_GOOD;
 }
 
 
 static void
-set_default_option (SANE_Handle handle, SANE_Int option_index)
+set_default_option (SANE_Handle handle, const SANE_Option_Descriptor *option, SANE_Int option_index)
 {
     SANE_Status status;
     
@@ -416,7 +432,7 @@ set_default_option (SANE_Handle handle, SANE_Int option_index)
     g_debug ("sane_control_option (%d, SANE_ACTION_SET_AUTO) -> %s",
              option_index, get_status_string (status));
     if (status != SANE_STATUS_GOOD)
-        g_warning ("Error setting control option: %s", sane_strstatus(status));
+        g_warning ("Error setting default option %s: %s", option->name, sane_strstatus(status));
 }
 
 
@@ -425,7 +441,7 @@ set_bool_option (SANE_Handle handle, const SANE_Option_Descriptor *option, SANE_
 {
     SANE_Bool v = value;
     g_return_if_fail (option->type == SANE_TYPE_BOOL);
-    control_option (handle, SANE_TYPE_BOOL, option_index, SANE_ACTION_SET_VALUE, &v);
+    control_option (handle, option, option_index, SANE_ACTION_SET_VALUE, &v);
 }
 
 
@@ -444,18 +460,68 @@ set_int_option (SANE_Handle handle, const SANE_Option_Descriptor *option, SANE_I
         if (v > option->constraint.range->max)
             v = option->constraint.range->max;
     }
-    control_option (handle, SANE_TYPE_INT, option_index, SANE_ACTION_SET_VALUE, &v);
+    else if (option->constraint_type == SANE_CONSTRAINT_WORD_LIST) {
+        int i;
+        SANE_Int min = INT_MIN, max = INT_MAX;
+      
+        /* Find nearest value above and below requested */
+        for (i = 0; i < option->constraint.word_list[0]; i++) {
+            SANE_Int x = option->constraint.word_list[i+1];
+            if (x <= v && x > min)
+                min = x;
+            if (x >= v && x < max)
+                max = x;
+        }
+      
+        /* Pick nearest */
+        if (max - v < v - min)
+            v = max;
+        else
+            v = min;
+    }
+
+    control_option (handle, option, option_index, SANE_ACTION_SET_VALUE, &v);
 }
 
 
 static void
 set_fixed_option (SANE_Handle handle, const SANE_Option_Descriptor *option, SANE_Int option_index, double value)
 {
-    SANE_Fixed v = SANE_FIX (value);
+    SANE_Fixed v;
 
     g_return_if_fail (option->type == SANE_TYPE_FIXED);
 
-    control_option (handle, SANE_TYPE_FIXED, option_index, SANE_ACTION_SET_VALUE, &v);
+    if (option->constraint_type == SANE_CONSTRAINT_RANGE) {
+        double min = SANE_UNFIX (option->constraint.range->min);
+        double max = SANE_UNFIX (option->constraint.range->max);
+
+        if (value < min)
+            value = min;
+        if (value > max)
+            value = max;
+    }
+    else if (option->constraint_type == SANE_CONSTRAINT_WORD_LIST) {
+        int i;
+        double min = DBL_MIN, max = DBL_MAX;
+      
+        /* Find nearest value above and below requested */
+        for (i = 0; i < option->constraint.word_list[0]; i++) {
+            double x = SANE_UNFIX (option->constraint.word_list[i+1]);
+            if (x <= value && x > min)
+                min = x;
+            if (x >= value && x < max)
+                max = x;
+        }
+      
+        /* Pick nearest */
+        if (max - value < value - min)
+            value = max;
+        else
+            value = min;
+    }
+
+    v = SANE_FIX (value);
+    control_option (handle, option, option_index, SANE_ACTION_SET_VALUE, &v);
 }
 
 
@@ -472,7 +538,7 @@ set_string_option (SANE_Handle handle, const SANE_Option_Descriptor *option, SAN
     size = option->size > value_size ? option->size : value_size;
     string = g_malloc(sizeof(char) * size);
     strcpy (string, value);
-    result = control_option (handle, SANE_TYPE_STRING, option_index, SANE_ACTION_SET_VALUE, string);
+    result = control_option (handle, option, option_index, SANE_ACTION_SET_VALUE, string);
     g_free (string);
    
     return result;
@@ -586,7 +652,10 @@ log_option (SANE_Int index, const SANE_Option_Descriptor *option)
         for (i = 0; i < option->constraint.word_list[0]; i++) {
             if (i != 0)
                 g_string_append (string, ", ");
-            g_string_append_printf (string, "%d", option->constraint.word_list[i+1]);
+            if (option->type == SANE_TYPE_INT)
+                g_string_append_printf (string, "%d", option->constraint.word_list[i+1]);
+            else
+                g_string_append_printf (string, "%f", SANE_UNFIX (option->constraint.word_list[i+1]));
         }
         g_string_append (string, "]");
         break;
@@ -699,6 +768,16 @@ close_device (Scanner *scanner)
     }
     g_list_free (scanner->priv->job_queue);
     scanner->priv->job_queue = NULL;
+  
+    set_scanning (scanner, FALSE);
+}
+
+static void
+fail_scan (Scanner *scanner, gint error_code, const gchar *error_string)
+{
+    close_device (scanner);
+    scanner->priv->state = STATE_IDLE;
+    emit_signal (scanner, SCAN_FAILED, g_error_new (SCANNER_TYPE, error_code, "%s", error_string));
 }
 
 
@@ -730,15 +809,10 @@ handle_requests (Scanner *scanner)
 
          case REQUEST_START_SCAN:
              scanner->priv->job_queue = g_list_append (scanner->priv->job_queue, request->job);
-             if (g_list_length (scanner->priv->job_queue) == 1)
-                 scanner->priv->state = STATE_OPEN;
              break;
 
          case REQUEST_CANCEL:
-             close_device (scanner);
-             scanner->priv->state = STATE_IDLE;
-             emit_signal (scanner, SCAN_FAILED,
-                          g_error_new (SCANNER_TYPE, SANE_STATUS_CANCELLED, "Scan cancelled - do not report this error"));
+             fail_scan (scanner, SANE_STATUS_CANCELLED, "Scan cancelled - do not report this error");
              break;
 
          case REQUEST_QUIT:
@@ -768,17 +842,14 @@ do_open (Scanner *scanner)
     scanner->priv->notified_page = -1;
     scanner->priv->option_index = 0;
 
-    if (!job->device)
-        job->device = g_strdup (scanner->priv->default_device->name);
+    if (!job->device && scanner->priv->default_device)
+        job->device = g_strdup (scanner->priv->default_device);
 
     if (!job->device) {
-        scanner->priv->state = STATE_IDLE;
         g_warning ("No scan device available");
-        scanner->priv->state = STATE_IDLE;
-        emit_signal (scanner, SCAN_FAILED,
-                     g_error_new (SCANNER_TYPE, status,
-                                  /* Error displayed when no scanners to scan with */
-                                  _("No scanners available.  Please connect a scanner.")));
+        fail_scan (scanner, status,
+                   /* Error displayed when no scanners to scan with */
+                   _("No scanners available.  Please connect a scanner."));
         return;
     }
  
@@ -796,19 +867,16 @@ do_open (Scanner *scanner)
   
     g_free (scanner->priv->current_device);
     scanner->priv->current_device = NULL;
-  
+
     status = sane_open (job->device, &scanner->priv->handle);
     g_debug ("sane_open (\"%s\") -> %s", job->device, get_status_string (status));
 
     if (status != SANE_STATUS_GOOD) {
         g_warning ("Unable to get open device: %s", sane_strstatus (status));
         scanner->priv->handle = NULL;
-        close_device (scanner);
-        scanner->priv->state = STATE_IDLE;
-        emit_signal (scanner, SCAN_FAILED,
-                     g_error_new (SCANNER_TYPE, status,
-                                  /* Error displayed when cannot connect to scanner */
-                                  _("Unable to connect to scanner")));
+        fail_scan (scanner, status,
+                   /* Error displayed when cannot connect to scanner */
+                   _("Unable to connect to scanner"));
         return;
     }
 
@@ -837,14 +905,28 @@ do_get_option (Scanner *scanner)
     }
 
     log_option (option_index, option);
-    if (!option->name)
+  
+    /* Ignore groups */
+    if (option->type == SANE_TYPE_GROUP)
+        return;
+
+    /* Option disabled */
+    if (option->cap & SANE_CAP_INACTIVE)
+        return;
+  
+    /* Some options are unnammed (e.g. Option 0) */
+    if (option->name == NULL)
         return;
 
     if (strcmp (option->name, SANE_NAME_SCAN_RESOLUTION) == 0) {
-        if (option->type == SANE_TYPE_FIXED)
-            set_fixed_option (scanner->priv->handle, option, option_index, job->dpi);
-        else
-            set_int_option (scanner->priv->handle, option, option_index, job->dpi);
+        if (option->type == SANE_TYPE_FIXED) {
+            double dpi = job->dpi;
+            set_fixed_option (scanner->priv->handle, option, option_index, dpi);
+        }
+        else {
+            SANE_Int dpi = job->dpi;
+            set_int_option (scanner->priv->handle, option, option_index, dpi);
+        }
     }
     else if (strcmp (option->name, SANE_NAME_SCAN_SOURCE) == 0) {
         const char *adf_sources[] =
@@ -862,7 +944,7 @@ do_get_option (Scanner *scanner)
                  g_warning ("Unable to set ADF source, please file a bug");
         }
         else {
-            set_default_option (scanner->priv->handle, option_index);
+            set_default_option (scanner->priv->handle, option, option_index);
         }
     }
     else if (strcmp (option->name, SANE_NAME_BIT_DEPTH) == 0) {
@@ -993,6 +1075,7 @@ do_complete_document (Scanner *scanner)
     // TODO
 
     emit_signal (scanner, DOCUMENT_DONE, NULL);
+    set_scanning (scanner, FALSE);
 }
 
 
@@ -1000,7 +1083,9 @@ static void
 do_start (Scanner *scanner)
 {
     SANE_Status status;
-  
+
+    emit_signal (scanner, EXPECT_PAGE, NULL);
+
     status = sane_start (scanner->priv->handle);
     g_debug ("sane_start (page=%d, pass=%d) -> %s", scanner->priv->page_number, scanner->priv->pass_number, get_status_string (status));
     if (status == SANE_STATUS_GOOD) {
@@ -1011,12 +1096,9 @@ do_start (Scanner *scanner)
     }
     else {
         g_warning ("Unable to start device: %s", sane_strstatus (status));
-        close_device (scanner);
-        scanner->priv->state = STATE_IDLE;
-        emit_signal (scanner, SCAN_FAILED,
-                     g_error_new (SCANNER_TYPE, status,
-                                  /* Error display when unable to start scan */
-                                  _("Unable to start scan")));
+        fail_scan (scanner, status,
+                   /* Error display when unable to start scan */
+                   _("Unable to start scan"));
     }
 }
 
@@ -1032,12 +1114,9 @@ do_get_parameters (Scanner *scanner)
     g_debug ("sane_get_parameters () -> %s", get_status_string (status));
     if (status != SANE_STATUS_GOOD) {
         g_warning ("Unable to get device parameters: %s", sane_strstatus (status));
-        close_device (scanner);
-        scanner->priv->state = STATE_IDLE;
-        emit_signal (scanner, SCAN_FAILED,
-                     g_error_new (SCANNER_TYPE, status,
-                                  /* Error displayed when communication with scanner broken */
-                                  _("Error communicating with scanner")));
+        fail_scan (scanner, status,
+                   /* Error displayed when communication with scanner broken */
+                   _("Error communicating with scanner"));
         return;
     }
   
@@ -1063,8 +1142,8 @@ do_get_parameters (Scanner *scanner)
     }
 
     /* Prepare for read */
-    scanner->priv->bytes_remaining = scanner->priv->parameters.bytes_per_line;
-    scanner->priv->buffer = g_malloc(sizeof(SANE_Byte) * scanner->priv->bytes_remaining);
+    scanner->priv->buffer = g_malloc(sizeof(SANE_Byte) * MAX_READ);
+    scanner->priv->n_used = 0;
     scanner->priv->line_count = 0;
     scanner->priv->pass_number = 0;
     scanner->priv->state = STATE_READ;
@@ -1075,6 +1154,8 @@ static void
 do_complete_page (Scanner *scanner)
 {
     ScanJob *job;
+
+    emit_signal (scanner, PAGE_DONE, NULL);
   
     job = (ScanJob *) scanner->priv->job_queue->data;
 
@@ -1097,7 +1178,6 @@ do_complete_page (Scanner *scanner)
         return;
     }
 
-    emit_signal (scanner, PAGE_DONE, NULL);  
     do_complete_document (scanner);
 }
 
@@ -1106,13 +1186,15 @@ static void
 do_read (Scanner *scanner)
 {
     SANE_Status status;
-    SANE_Int n_read;
+    SANE_Int n_to_read, n_read;
+
+    /* Read as many bytes as we expect */
+    n_to_read = MAX_READ - scanner->priv->n_used;
 
     status = sane_read (scanner->priv->handle,
-                        scanner->priv->buffer + (scanner->priv->parameters.bytes_per_line - scanner->priv->bytes_remaining),
-                        scanner->priv->bytes_remaining,
-                        &n_read);
-    g_debug ("sane_read (%d) -> (%s, %d)", scanner->priv->bytes_remaining, get_status_string (status), n_read);
+                        scanner->priv->buffer + scanner->priv->n_used,
+                        n_to_read, &n_read);
+    g_debug ("sane_read (%d) -> (%s, %d)", n_to_read, get_status_string (status), n_read);
 
     /* End of variable length frame */
     if (status == SANE_STATUS_EOF &&
@@ -1125,17 +1207,16 @@ do_read (Scanner *scanner)
     /* Communication error */
     if (status != SANE_STATUS_GOOD) {
         g_warning ("Unable to read frame from device: %s", sane_strstatus (status));
-        close_device (scanner);
-        scanner->priv->state = STATE_IDLE;
-        emit_signal (scanner, SCAN_FAILED,
-                     g_error_new (SCANNER_TYPE, status,
-                                  /* Error displayed when communication with scanner broken */
-                                  _("Error communicating with scanner")));
+        fail_scan (scanner, status,
+                   /* Error displayed when communication with scanner broken */
+                   _("Error communicating with scanner"));
         return;
     }
  
-    scanner->priv->bytes_remaining -= n_read;
-    if (scanner->priv->bytes_remaining == 0) {
+    scanner->priv->n_used += n_read;
+  
+    /* Feed out lines */
+    if (scanner->priv->n_used >= scanner->priv->parameters.bytes_per_line) {
         ScanLine *line;
 
         line = g_malloc(sizeof(ScanLine));
@@ -1161,17 +1242,27 @@ do_read (Scanner *scanner)
         line->data = scanner->priv->buffer;
         line->data_length = scanner->priv->parameters.bytes_per_line;
         line->number = scanner->priv->line_count;
-        emit_signal (scanner, GOT_LINE, line);
+        line->n_lines = scanner->priv->n_used / line->data_length;
+
         scanner->priv->buffer = NULL;
+        scanner->priv->line_count += line->n_lines;
 
         /* On last line */
-        scanner->priv->line_count++;
-        if (scanner->priv->parameters.lines > 0 && scanner->priv->line_count == scanner->priv->parameters.lines) {
+        if (scanner->priv->parameters.lines > 0 && scanner->priv->line_count >= scanner->priv->parameters.lines) {
+            emit_signal (scanner, GOT_LINE, line);
             do_complete_page (scanner);
         }
         else {
-            scanner->priv->bytes_remaining = scanner->priv->parameters.bytes_per_line;
-            scanner->priv->buffer = g_malloc(sizeof(SANE_Byte) * scanner->priv->bytes_remaining);
+            int i, n_remaining;
+
+            scanner->priv->buffer = g_malloc(sizeof(SANE_Byte) * MAX_READ);
+            n_remaining = scanner->priv->n_used - (line->n_lines * line->data_length);
+            scanner->priv->n_used = 0;
+            for (i = 0; i < n_remaining; i++) {
+                scanner->priv->buffer[i] = line->data[i + (line->n_lines * line->data_length)];
+                scanner->priv->n_used++;                
+            }
+            emit_signal (scanner, GOT_LINE, line);
         }
     }
 }
@@ -1204,8 +1295,10 @@ scan_thread (Scanner *scanner)
     while (handle_requests (scanner)) {
         switch (scanner->priv->state) {
         case STATE_IDLE:
-             if (scanner->priv->job_queue)
+             if (scanner->priv->job_queue) {
+                 set_scanning (scanner, TRUE);
                  scanner->priv->state = STATE_OPEN;
+             }
              break;
         case STATE_REDETECT:
             do_redetect (scanner);
@@ -1291,7 +1384,6 @@ scanner_scan (Scanner *scanner, const char *device,
     request->job->depth = depth;
     request->job->multi_page = multi_page;
     g_async_queue_push (scanner->priv->scan_queue, request);
-    scanner->priv->scanning = TRUE;
 }
 
 
@@ -1346,6 +1438,14 @@ scanner_class_init (ScannerClass *klass)
                       NULL, NULL,
                       g_cclosure_marshal_VOID__POINTER,
                       G_TYPE_NONE, 1, G_TYPE_POINTER);
+    signals[EXPECT_PAGE] =
+        g_signal_new ("expect-page",
+                      G_TYPE_FROM_CLASS (klass),
+                      G_SIGNAL_RUN_LAST,
+                      G_STRUCT_OFFSET (ScannerClass, expect_page),
+                      NULL, NULL,
+                      g_cclosure_marshal_VOID__VOID,
+                      G_TYPE_NONE, 0);
     signals[GOT_PAGE_INFO] =
         g_signal_new ("got-page-info",
                       G_TYPE_FROM_CLASS (klass),
@@ -1386,6 +1486,14 @@ scanner_class_init (ScannerClass *klass)
                       NULL, NULL,
                       g_cclosure_marshal_VOID__VOID,
                       G_TYPE_NONE, 0);
+    signals[SCANNING_CHANGED] =
+        g_signal_new ("scanning-changed",
+                      G_TYPE_FROM_CLASS (klass),
+                      G_SIGNAL_RUN_LAST,
+                      G_STRUCT_OFFSET (ScannerClass, scanning_changed),
+                      NULL, NULL,
+                      g_cclosure_marshal_VOID__VOID,
+                      G_TYPE_NONE, 0);
 
     g_type_class_add_private (klass, sizeof (ScannerPrivate));
 
