@@ -34,6 +34,9 @@ typedef struct
 {
     /* Line number */
     gint number;
+  
+    /* Number of lines in this packet */
+    gint n_lines;
 
     /* Width in pixels and format */
     gint width, depth;
@@ -74,11 +77,13 @@ typedef struct
 
     void (*update_devices) (Scanner *scanner, GList *devices);
     void (*authorize) (Scanner *scanner, const gchar *resource);
+    void (*expect_page) (Scanner *scanner);
     void (*got_page_info) (Scanner *scanner, ScanPageInfo *info);
     void (*got_line) (Scanner *scanner, ScanLine *line);
     void (*scan_failed) (Scanner *scanner, GError *error);
     void (*page_done) (Scanner *scanner);
     void (*document_done) (Scanner *scanner);
+    void (*scanning_changed) (Scanner *scanner);
 } ScannerClass;
 
 
